"""Top-level package for ncem."""

import ncem.estimators
import ncem.interpretation
import ncem.models
import ncem.train

__maintainer__ = ", ".join(["Anna C. Schaar", "David S. Fischer"])
__author__ = ", ".join(["Anna C. Schaar", "David S. Fischer"])
__email__ = ", ".join(["anna.schaar@helmholtz-muenchen.de", "david.fischer@helmholtz-muenchen.de"])
__version__ = "0.1.0"
